classdef Experiment
    % EXPERIMENT  Runs a fully-decentralised federated learning experiment.
    %
    % An experiment consists of distributing a dataset amongst several nodes, and
    % then, in each round, selecting a node which learns from their neighbours
    % before training their own model a bit, continuing rounds until the nodes'
    % models have converged.
    %
    % See also: ExperimentConfig, PerformanceSim

    properties
        % IDX  This [Experiment]'s unique identifier, containing both its identity and
        % its repetition number.
        idx (1, 2) {mustBeInteger, mustBePositive} = [1, 1];

        % CONF  The configuration on how to run the [Experiment].
        conf (1, 1) ExperimentConfig;

        % RAW_GRAPH  The created graph before stretching.
        raw_graph (1, 1) graph;
        % GRAPH  The created graph in which users are configured.
        graph (1, 1) graph;
        % MODELS  The trained (or to-be-trained) models.
        models (:, 1) cell;  % cell<Model>

        % ROUND_IDX  The number of rounds executed.
        round_idx (1, 1) = missing;
        % CONVERGENCE  The convergence metric that has been measured.
        convergence (1, 1) = missing;
    end


    methods
        function obj = Experiment(idx, conf)
            % EXPERIMENT  Constructs experiment number [idx] using [config].

            arguments% (Input)
                idx (1, 2) {mustBeInteger, mustBePositive};
                conf (1, 1) ExperimentConfig;
            end
            % arguments (Output)
            %     obj (1, 1) Experiment;
            % end

            obj.idx = idx;
            obj.conf = conf;
        end


        function obj = run(obj)
            % RUN  Runs this experiment.

            %% Create network graph
            attempt_idx = 0;
            while true
                attempt_idx = attempt_idx + 1;
                if attempt_idx > obj.conf.network_max_attempts
                    error("Failed to generate network in maximum number of attempts.");
                end

                if obj.conf.network_layout == "erdos-renyi"
                    obj.raw_graph = Graphs.generate_erdos_renyi(obj.conf.node_count, obj.conf.network_erdos_renyi_p);
                elseif obj.conf.network_layout == "watts-strogatz"
                    obj.raw_graph = Graphs.generate_watts_strogatz(obj.conf.node_count, ...
                                                                   obj.conf.network_watts_strogatz_k, ...
                                                                   obj.conf.network_watts_strogatz_p);
                elseif obj.conf.network_layout == "barabasi-albert"
                    obj.raw_graph = Graphs.generate_barabasi_albert(obj.conf.node_count, obj.conf.network_barabasi_albert_m);
                elseif obj.conf.network_layout == "complete"
                    obj.raw_graph = Graphs.generate_complete(obj.conf.node_count);
                elseif obj.conf.network_layout == "empty"
                    obj.raw_graph = Graphs.generate_empty(obj.conf.node_count);
                else
                    error("Unknown network layout '%s'.", obj.conf.network_layout);
                end

                obj.graph = Graphs.stretched(obj.raw_graph, ...
                                             min_girth = obj.conf.network_min_girth, ...
                                             remove = obj.conf.network_stretch_remove, ...
                                             place = obj.conf.network_stretch_place);

                if obj.conf.network_require_connected && max(conncomp(obj.graph)) == 1; break; end
            end


            %% Run federated averaging
            % Initialize models
            dataset = 50 * rand([obj.conf.node_count, 1]);
            obj.models = cell(obj.conf.node_count, 1);
            for selected_node = 1:obj.conf.node_count
                obj.models{selected_node} = Model(dataset(selected_node));
            end

            % Run rounds
            obj.round_idx = 0;
            while true
                if obj.conf.fl_rounds >= 0 && obj.round_idx >= obj.conf.fl_rounds
                    break;
                end
                if obj.conf.fl_convergence_threshold >= 0 && obj.convergence <= obj.conf.fl_convergence_threshold
                    break;
                end

                obj.round_idx = obj.round_idx + 1;

                % Select node
                selected_node = randsample(obj.conf.node_count, 1);

                % Average with neighbors
                obj.models{selected_node}.combine_with_neighbors(obj.conf.fl_self_weight, ...
                                                                 obj.models(neighbors(obj.graph, selected_node)));

                % Validation
                if obj.conf.fl_convergence_threshold >= 0
                    obj.convergence = Model.calc_convergence_distance(obj.models);
                end
            end
        end


        function fig = plot_stretching(obj)
            % PLOT_STRETCHING  Visualizes how this experiment's graph got stretched.

            fig = figure;
            sgtitle(sprintf("Stretching to girth %d", obj.conf.network_min_girth));

            P1 = subplot(1, 3, 3);
            H = plot(obj.graph, Layout = "force3");
            title(sprintf("Edges compensated"));

            P2 = subplot(1, 3, 2);
            plot(Graphs.stretched(obj.raw_graph, min_girth = obj.conf.network_min_girth, ...
                                  remove = "any", place = "none"), ...
                 XData = H.XData, YData = H.YData, ZData = H.ZData);
            title("Edges removed");

            P3 = subplot(1, 3, 1);
            plot(obj.raw_graph, XData = H.XData, YData = H.YData, ZData = H.ZData);
            title("Initial graph");

            setappdata(fig, MyCameraLink = linkprop([P1, P2, P3], ...
                                                    {"CameraUpVector", "CameraPosition", "CameraTarget", ...
                                                     "XLim", "YLim", "ZLim"}));
        end
    end
end
