classdef ExperimentConfig < Config
    % EXPERIMENTCONFIG  Configuration for running an [Experiment].

    properties
        % NODE_COUNT  Number of nodes in the network.
        node_count (1, 1) {mustBeInteger, mustBePositive} = 60;

        % NETWORK_LAYOUT  Graph type to use for the network layout.
        %
        % Must be one of: "erdos-renyi", "watts-strogatz", "barabasi-albert",
        % "complete", "empty".
        network_layout (1, 1) {mustBeMember(network_layout, ["erdos-renyi", ...
                                                             "watts-strogatz", ...
                                                             "barabasi-albert", ...
                                                             "complete", ...
                                                             "empty"])} = "erdos-renyi";
        % NETWORK_ERDOS_RENYI_P  The `p` parameter for Erdős-Renyi graphs.
        network_erdos_renyi_p (1, 1) {mustBeFloat, mustBeInRange(network_erdos_renyi_p, 0, 1)} = 0.1;
        % NETWORK_WATTS_STROGATZ_K  The `k` parameter for Watts-Strogatz graphs.
        network_watts_strogatz_k (1, 1) {mustBeFloat, mustBeNonnegative} = 12;
        % NETWORK_WATTS_STROGATZ_P  The `p` parameter for Watts-Strogatz graphs.
        network_watts_strogatz_p (1, 1) {mustBeFloat, mustBeInRange(network_watts_strogatz_p, 0, 1)} = 0.5;
        % NETWORK_BARABASI_ALBERT_M  The `m` parameter for Barabási-Albert graphs.
        network_barabasi_albert_m (1, 1) {mustBeInteger, mustBePositive} = 12;
        % NETWORK_MIN_GIRTH  The girth to which the graph should be stretched (at least)
        % in post-processing.
        %
        % Set this number to -1 if all cycles should be removed, or to 0 to not remove
        % any cycles.
        network_min_girth (1, 1) {mustBeInteger} = 0;
        % NETWORK_STRETCH_REMOVE  The method by which nodes to be removed should be
        % selected when stretching a graph to a particular girth. See also
        % [Graphs#stretched].
        network_stretch_remove (1, 1) {mustBeText, mustBeMember(network_stretch_remove, ["any", "basis"])} = "basis";
        % NETWORK_STRETCH_PLACE  The method by which nodes to be added should be
        % selected after removing a node while stretching a graph to a particular girth.
        % See also [Graphs#stretched].
        network_stretch_place (1, 1) {mustBeText, mustBeMember(network_stretch_place, ["none", "diameter"])} = "none";
        % NETWORK_REQUIRE_CONNECTED  `true` if and only if the created network must
        % consist of a single connected component.
        network_require_connected (1, 1) logical = true;
        % NETWORK_MAX_TRIALS  The maximum number of attempts to generate a network
        % before erroring.
        network_max_attempts (1, 1) {mustBeInteger, mustBePositive} = 250;

        % FL_SELF_WEIGHT  The weight of the node's own model relative to each other
        % model when averaging with neighbours.
        fl_self_weight (1, 1) = 1;
        % FL_ROUNDS  Number of federated learning rounds to run, or a negative number to
        % not limit the experiment by the number of rounds.
        fl_rounds (1, 1) {mustBeInteger} = 50;
        % FL_CONVERGENCE_THRESHOLD  Largest difference between model parameters before
        % convergence is achieved, or a negative number to not limit the experiment by
        % convergence.
        fl_convergence_threshold (1, 1) {mustBeFloat} = -1;
    end


    methods
        function obj = ExperimentConfig(args)
            arguments% (Input)
                args.?ExperimentConfig;
            end
            % arguments (Output)
            %     obj (1, 1) ExperimentConfig;
            % end

            obj = obj@Config(args);
        end


        function new_obj = set(obj, args)
            arguments% (Input)
                obj (1, 1) ExperimentConfig;
                args.?ExperimentConfig;
            end
            % arguments (Output)
            %     new_obj (1, 1) ExperimentConfig;
            % end

            new_obj = set@Config(obj, args);
        end

        function h = cache_id(obj)
            arguments% (Input)
                obj (1, 1) ExperimentConfig;
            end
            % arguments (Output)
            %     h (1, 1) {mustBeText};
            % end

            h = cache_id@Config(obj);
        end
    end

    methods (Static)
        function partial_obj = partial(args)
            arguments% (Input)
                args.?ExperimentConfig;
            end
            % arguments (Output)
            %     partial_obj (1, :) cell;
            % end

            partial_obj = partial@Config(args);
        end
    end
end
