classdef Model < handle
    % MODEL  A model represented by a single number.

    properties (Access = private)
        % NUMBER  This model's singular number.
        number (1, 1) {mustBeNumeric};
    end


    methods
        function obj = Model(number)
            % MODEL  Constructs a new model with the given number.

            obj.number = number;
        end


        function combine_with_neighbors(obj, self_weight, neighbor_models)
            % COMBINE_WITH_NEIGHBORS  Adjusts this model to be the weighted average of
            % itself (with weight [self_weight]) and each model in [neighbor_models] (each
            % with a weight of 1).

            arguments% (Input)
                obj (1, 1) Model;
                self_weight (1, 1) {mustBeInteger};
                neighbor_models (:, 1) cell;  % cell<Model>
            end

            if numel(neighbor_models) == 0; return; end

            obj.number = mean([self_weight * obj.number; cellfun(@(it) it.number, neighbor_models(:))]);
        end


        function accuracy = evaluate_accuracy(obj, dataset)
            % EVALUATE_ACCURACY  Returns the [accuracy] of this model, which is the
            % (absolute) distance to the mean of the original [dataset].

            arguments% (Input)
                obj (1, 1) Model;
                dataset (:, 1) {mustBeNumeric};
            end
            % arguments (Output)
            %     accuracy (1, 1) double {mustBeInRange(accuracy, 0, 1)};
            % end

            accuracy = abs(obj.number - mean(dataset));
        end
    end

    methods (Static)
        function distance = calc_convergence_distance(models)
            % CALC_CONVERGENCE_DISTANCE  Returns the largest pairwise difference between any
            % two [models].

            arguments% (Input)
                models (:, 1) cell {mustBeNonempty};  % cell<Model>
            end
            % arguments (Output)
            %     distance (1, 1) double;
            % end

            if isscalar(models)
                distance = 0;
            else
                all_numbers = cellfun(@(it) it.number', models);
                distance = max(max(all_numbers, [], 1) - min(all_numbers, [], 1));
            end
        end
    end
end
