clear;

part_count = str2double(getenv("SLURM_ARRAY_TASK_MAX"));
if isnan(part_count); part_count = 1; end
part_idx = str2double(getenv("SLURM_ARRAY_TASK_ID"));
if isnan(part_idx); part_idx = 0; end


%% Configure
% Laboratory
lab_conf = ...
    LaboratoryConfig.partial( ...
        seed = 17, ...
        load_enabled = true, ...
        ... load_exp_behavior = "skip", ...
        load_validate = false, ...
        save_enabled = true, ...
        ...
        parallel = true, ...
        parallel_max_workers = Inf, ...
        repeat_count = 1000, ...
        ...
        plot_show = true, ...
        plot_save = true, ...
        plot_save_formats = ["eps"], ...
        ...
        part_count = part_count, ...
        part_idx = part_idx ...
    );

% Experiments
exp_base_conf = ...
    ExperimentConfig( ...
        node_count = 50, ...
        fl_convergence_threshold = 1, ...
        fl_rounds = -1, ...
        network_stretch_remove = "any", ...
        network_stretch_place = "none" ...
    );
exp_confs = ...
    Config.combinations( ...
        { ...
            ExperimentConfig.partial(network_layout = "erdos-renyi", ...
                                     network_erdos_renyi_p = 0.1), ...
            ExperimentConfig.partial(network_layout = "erdos-renyi", ...
                                     network_erdos_renyi_p = 0.5), ...
            ExperimentConfig.partial(network_layout = "erdos-renyi", ...
                                     network_erdos_renyi_p = 0.9), ...
        }, ...
        arrayfun(@(it) ExperimentConfig.partial(network_min_girth = it), 0:25, UniformOutput = false) ...
    );

% Plots
plot_base_conf = PlotConfig();
plot_confs = { ...
    PlotConfig.partial( ...
        title = "Convergence by Girth", ...
        x_label = "Stretched Minimal Girth", ...
        y_label = "Global Rounds Until Convergence", ...
        exp_to_data = @(it) [girth_to_y(it.conf), it.round_idx], ...
        exp_to_line = @(it) hash(it.conf.set(network_min_girth = 1)), ...
        line_to_label = @(it) conf_to_label(it.conf), ...
        line_to_style = @(it) conf_to_style(it.conf, 2), ...
        y_sci_ticks = false ...
    ), ...
};


%% Run
exp_confs = num2cell(cellfun(@(it) exp_base_conf.set(it{:}), exp_confs));
plot_confs = num2cell(cellfun(@(it) plot_base_conf.set(it{:}), plot_confs));

lab = Laboratory(LaboratoryConfig(exp_confs = exp_confs, plot_confs = plot_confs).set(lab_conf{:}));
lab.run();
lab.plot();


%% Helpers
function string = conf_to_label(conf)
    if conf.network_layout == "erdos-renyi"
        string = sprintf("p=%.1f", conf.network_erdos_renyi_p);
    elseif conf.network_layout == "watts-strogatz"
        string = sprintf("watts-strogatz (p=%.1f)", conf.network_watts_strogatz_p);
    elseif conf.network_layout == "barabasi-albert"
        string = sprintf("barabasi-albert (m=%.0f)", conf.network_barabasi_albert_m);
    else
        error("Unknown network layout '%s'.", conf.network_layout);
    end
end

function style = conf_to_style(conf, width)
    arguments% (Input)
        conf (1, 1) ExperimentConfig;
        width (1, 1) {mustBeInteger} = 1;
    end

    if conf.network_layout == "erdos-renyi"
        style = {"LineStyle", "-"};
    elseif conf.network_layout == "watts-strogatz"
        style = {"LineStyle", "--"};
    elseif conf.network_layout == "barabasi-albert"
        style = {"LineStyle", ":"};
    else
        error("Unknown network layout '%s'.", conf.network_layout);
    end

    style = [style(:)', {"LineWidth"}, {width}];
end

function y = girth_to_y(conf)
    if conf.network_min_girth == -1
        y = conf.node_count;
    else
        y = conf.network_min_girth;
    end
end
